@@ -2,7 +2,8 @@
  *  802.11 WEP network connection tunneling
  *  based on aireplay-ng
  *
- *  Copyright (C) 2006 Martin Beck
+ *  Copyright (C) 2006,2007 Thomas d'Otreppe
+ *  Copyright (C) 2006,2007 Martin Beck
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -19,20 +20,41 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-#ifndef linux
-	#error Airtun-ng only compiles with Linux
+#if !(defined(linux) || defined(__FreeBSD__) || defined(__FreeBSD_kernel__))
+    #warning Airtun-ng could fail on this OS
+#endif
+
+#ifdef linux
+    #include <linux/rtc.h>
 #endif
 
-#include <linux/rtc.h>
 #include <sys/socket.h>
 #include <sys/ioctl.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <sys/time.h>
 
-#include <netpacket/packet.h>
-#include <linux/if_ether.h>
-#include <linux/wireless.h>
+#if defined(linux)
+    #include <netpacket/packet.h>
+    #include <linux/if_ether.h>
+    #include <linux/if.h>
+    #include <linux/wireless.h>
+#endif /* linux */
+
+#if defined(__FreeBSD__) || defined( __FreeBSD_kernel__)
+    #include <sys/param.h>
+    #include <sys/sysctl.h>
+    #include <sys/uio.h>
+    #include <net/bpf.h>
+    #include <net/if.h>
+    #include <net/if_media.h>
+    #include <netinet/in.h>
+    #include <netinet/if_ether.h>
+    #include <net80211/ieee80211.h>
+    #include <net80211/ieee80211_freebsd.h>
+    #include <net80211/ieee80211_radiotap.h>
+#endif /* __FreeBSD__ */
+
 #include <arpa/inet.h>
 #include <unistd.h>
 #include <dirent.h>
@@ -45,11 +67,16 @@
 #include <getopt.h>
 
 #include <fcntl.h>
-#include <errno.h>
-#include <time.h>
+//#include <errno.h>
+//#include <time.h>
 
-#include <linux/if.h>
-#include <linux/if_tun.h>
+#if defined(linux)
+    #include <linux/if_tun.h>
+#endif
+
+#if defined(__FreeBSD__) || defined( __FreeBSD_kernel__)
+    #include <net/if_tun.h>
+#endif
 
 #include "version.h"
 #include "pcap.h"
@@ -69,34 +96,47 @@
 #define CRYPT_NONE 0
 #define CRYPT_WEP  1
 
+#if defined(linux)
 #define MAX(x,y) ( (x)>(y) ? (x) : (y) )
+#endif
 
 extern char * getVersion(char * progname, int maj, int min, int submin, int svnrev);
-extern int is_ndiswrapper(const char * iface, const char * path);
 extern char * searchInside(const char * dir, const char * filename);
-extern char * wiToolsPath(const char * tool);
 extern unsigned char * getmac(char * macAddress, int strict, unsigned char * mac);
 extern int check_crc_buf( unsigned char *buf, int len );
 extern int add_crc32(unsigned char* data, int length);
 
+#if defined(linux)
+extern int is_ndiswrapper(const char * iface, const char * path);
+extern char * wiToolsPath(const char * tool);
+#endif /* linux */
+
 extern const unsigned long int crc_tbl[256];
 extern const unsigned char crc_chop_tbl[256][4];
 
 
 char usage[] =
 "\n"
-"  %s - (C) 2006 Thomas d'Otreppe\n"
+"  %s - (C) 2006,2007 Thomas d'Otreppe\n"
 "  Original work: Christophe Devine and Martin Beck\n"
 "  http://www.aircrack-ng.org\n"
 "\n"
 "  usage: airtun-ng <options> <replay interface>\n"
 "\n"
-"      -x nbpps  : maximum number of packets per second\n"
-"      -a bssid  : set Access Point MAC address\n"
-"      -i iface  : capture packets from this interface\n"
-"      -y file   : read PRGA from this file\n"
-"      -w wepkey : use this WEP-KEY to encrypt packets\n"
-"      -t tods   : send frames to AP (1) or to client (0)\n"
+"      -x nbpps         : number of packets per second (default: 100)\n"
+"      -a bssid         : set Access Point MAC address\n"
+"      -i iface         : capture packets from this interface\n"
+"      -y file          : read PRGA from this file\n"
+"      -w wepkey        : use this WEP-KEY to encrypt packets\n"
+"      -t tods          : send frames to AP (1) or to client (0)\n"
+"      -r file          : read frames out of pcap file\n"
+"\n"
+"  Repeater options:\n"
+"      --repeat         : activates repeat mode\n"
+"      --bssid <mac>    : BSSID to repeat\n"
+"      --netmask <mask> : netmask for BSSID filter\n"
+"\n"
+"      --help           : Displays this usage screen\n"
 "\n";
 
 struct options
@@ -105,7 +145,11 @@ struct options
     unsigned char r_dmac[6];
     unsigned char r_smac[6];
 
+    unsigned char f_bssid[6];
+    unsigned char f_netmask[6];
+
     char *s_face;
+    char *s_file;
     uchar *prga;
 
     int r_nbpps;
@@ -114,6 +158,8 @@ struct options
 
     uchar wepkey[64];
     int weplen, crypt;
+
+    int repeat;
 }
 opt;
 
@@ -162,6 +208,26 @@ void sighandler( int signum )
         alarmed++;
 }
 
+int is_filtered_netmask(uchar *bssid)
+{
+    uchar mac1[6];
+    uchar mac2[6];
+    int i;
+
+    for(i=0; i<6; i++)
+    {
+        mac1[i] = bssid[i]     & opt.f_netmask[i];
+        mac2[i] = opt.f_bssid[i] & opt.f_netmask[i];
+    }
+
+    if( memcmp(mac1, mac2, 6) != 0 )
+    {
+        return( 1 );
+    }
+
+    return 0;
+}
+
 /* wlanng-aware frame sending routing */
 
 int send_packet( void *buf, size_t count )
@@ -172,7 +238,7 @@ int send_packet( void *buf, size_t count )
     {
         /* for some reason, wlan-ng requires a special header */
 
-        if( ( ((unsigned char *) buf)[0] & 3 ) != 3 )
+        if( ( ((unsigned char *) buf)[1] & 3 ) != 3 )
         {
             memcpy( tmpbuf, buf, 24 );
             memset( tmpbuf + 24, 0, 22 );
@@ -214,11 +280,24 @@ int send_packet( void *buf, size_t count )
 
     ret = write( dev.fd_out, buf, count );
 
+    if( ( dev.is_wlanng || dev.is_hostap ) &&
+        ( ((uchar *) buf)[1] & 3 ) == 2 )
+    {
+        unsigned char maddr[6];
+
+        /* Prism2 firmware swaps the dmac and smac in FromDS packets */
+
+        memcpy( maddr, buf + 4, 6 );
+        memcpy( buf + 4, buf + 16, 6 );
+        memcpy( buf + 16, maddr, 6 );
+    }
+
     if( ret < 0 )
     {
         if( errno == EAGAIN || errno == EWOULDBLOCK ||
             errno == ENOBUFS )
         {
+            perror("again");
             usleep( 10000 );
             return( 0 );
         }
@@ -235,6 +314,7 @@ int send_packet( void *buf, size_t count )
 
 int read_packet( void *buf, size_t count )
 {
+    unsigned char bssid[6];
     int caplen, n = 0;
 
     if( ( caplen = read( dev.fd_in, tmpbuf, count ) ) < 0 )
@@ -278,10 +358,81 @@ int read_packet( void *buf, size_t count )
 
     memcpy( buf, tmpbuf + n, caplen );
 
+    if( opt.repeat )
+    {
+        if( memcmp(opt.f_bssid, NULL_MAC, 6) != 0 )
+        {
+            switch( tmpbuf[1+n] & 3 )
+            {
+                case  0: memcpy( bssid, tmpbuf + n + 16, 6 ); break;
+                case  1: memcpy( bssid, tmpbuf + n +  4, 6 ); break;
+                case  2: memcpy( bssid, tmpbuf + n + 10, 6 ); break;
+                default: memcpy( bssid, tmpbuf + n +  4, 6 ); break;
+            }
+            if( memcmp(opt.f_netmask, NULL_MAC, 6) != 0 )
+            {
+                if(is_filtered_netmask(bssid)) return( caplen );
+            }
+            else
+            {
+                if( memcmp(opt.f_bssid, bssid, 6) != 0 ) return( caplen );
+            }
+        }
+        send_packet(buf, caplen);
+    }
+
     return( caplen );
 }
 
 
+int msleep( int msec )
+{
+    struct timeval tv, tv2;
+    float f, ticks;
+    int n;
+
+    if(msec == 0) msec = 1;
+
+    ticks = 0;
+
+    while( 1 )
+    {
+        /* wait for the next timer interrupt, or sleep */
+
+        if( dev.fd_rtc >= 0 )
+        {
+            if( read( dev.fd_rtc, &n, sizeof( n ) ) < 0 )
+            {
+                perror( "read(/dev/rtc) failed" );
+                return( 1 );
+            }
+
+            ticks++;
+        }
+        else
+        {
+            /* we can't trust usleep, since it depends on the HZ */
+
+            gettimeofday( &tv,  NULL );
+            usleep( 1024 );
+            gettimeofday( &tv2, NULL );
+
+            f = 1000000 * (float) ( tv2.tv_sec  - tv.tv_sec  )
+                        + (float) ( tv2.tv_usec - tv.tv_usec );
+
+            ticks += f / 1024;
+        }
+
+        if( ( ticks / 1024 * 1000 ) < msec )
+            continue;
+
+        /* threshold reached */
+        break;
+    }
+
+    return 0;
+}
+
 #define PCT { struct tm *lt; time_t tc = time( NULL ); \
               lt = localtime( &tc ); printf( "%02d:%02d:%02d  ", \
               lt->tm_hour, lt->tm_min, lt->tm_sec ); }
@@ -359,8 +510,6 @@ int xor_keystream(uchar *ph80211, uchar *keystream, int len)
     return 0;
 }
 
-// ### ### BEGIN OWN CODE ### ###
-
 void print_packet ( uchar h80211[], int caplen )
 {
 	int i,j;
@@ -579,6 +728,7 @@ int packet_xmit(uchar* packet, int length)
 
 int packet_recv(uchar* packet, int length)
 {
+	int unused;
     uchar K[64];
     uchar bssid[6];
 
@@ -653,7 +803,7 @@ int packet_recv(uchar* packet, int length)
         memcpy( h80211+14, packet+z+8, length-z-8);
         length = length -z-8+14;
 
-        write(dev.fd_tap, h80211, length);
+        unused = write(dev.fd_tap, h80211, length);
     }
     else
     {
@@ -663,8 +813,6 @@ int packet_recv(uchar* packet, int length)
     return 0;
 }
 
-// ### ### END OWN CODE ### ###
-
 int sysfs_inject=0;
 int opensysfs( char *iface, int fd) {
     int fd2;
@@ -682,6 +830,7 @@ int opensysfs( char *iface, int fd) {
     return 0;
 }
 
+#if defined(linux)
 /* interface initialization routine */
 
 int openraw( char *iface, int fd, int *arptype )
@@ -766,23 +915,74 @@ int openraw( char *iface, int fd, int *arptype )
 
     return( 0 );
 }
+#endif /* linux */
+
+#if defined(__FreeBSD__) || defined( __FreeBSD_kernel__)
+int openraw(char *name, int fd, int *arptype) {
+    int i;
+//    int fd = -1;
+    struct ifreq ifr;
+
+/*    for(i = 0;i < 10; i++) {
+        sprintf(buf, "/dev/bpf%d", i);
+
+        fd = open(buf, O_RDWR);
+
+        if(fd < 0) {
+            if(errno != EBUSY) {
+                perror("can't open /dev/bpf");
+                return(1);
+            }
+            continue;
+        }
+        else
+            break;
+    }
+
+    if(fd < 0) {
+        perror("can't open /dev/bpf");
+        return(1);
+    }
+*/
+    strncpy(ifr.ifr_name, name, sizeof(ifr.ifr_name)-1);
+    ifr.ifr_name[sizeof(ifr.ifr_name)-1] = 0;
+
+    if(ioctl(fd, BIOCSETIF, &ifr) < 0) {
+        perror("ioctl(BIOCSETIF)");
+        return(1);
+    }
+
+    i = 1;
+    if(ioctl(fd, BIOCIMMEDIATE, &i) < 0) {
+        perror("ioctl(BIOCIMMEDIATE)");
+        return(1);
+    }
+
+    *arptype = ifr.ifr_addr.sa_family;
+
+    return 0;
+}
+#endif /* __FreeBSD__ */
 
 char athXraw[] = "athXraw";
 
 int main( int argc, char *argv[] )
 {
-    int ret_val, len, i, n;
+    int ret_val, len, i, n, ret, unused;
     struct ifreq if_request;
+    struct pcap_pkthdr pkh;
     fd_set read_fds;
     unsigned char buffer[4096];
+    unsigned char bssid[6];
     char *s, buf[128];
+    int caplen;
 
     /* check the arguments */
 
     memset( &opt, 0, sizeof( opt ) );
     memset( &dev, 0, sizeof( dev ) );
 
-    opt.r_nbpps = 200;
+    opt.r_nbpps = 100;
     opt.tods    = 0;
 
     srand( time( NULL ) );
@@ -792,11 +992,15 @@ int main( int argc, char *argv[] )
         int option_index = 0;
 
         static struct option long_options[] = {
-            {0,             0, 0,  0 }
+            {"netmask", 1, 0, 'm'},
+            {"bssid",   1, 0, 'd'},
+            {"repeat",  0, 0, 'f'},
+            {"help",  0, 0, 'H'},
+            {0,         0, 0,  0 }
         };
 
         int option = getopt_long( argc, argv,
-                        "x:a:h:i:r:y:t:w:",
+                        "x:a:h:i:r:y:t:w:m:d:fH",
                         long_options, &option_index );
 
         if( option < 0 ) break;
@@ -807,12 +1011,23 @@ int main( int argc, char *argv[] )
 
                 break;
 
+            case ':' :
+
+	    		printf("\"%s --help\" for help.\n", argv[0]);
+                return( 1 );
+
+            case '?' :
+
+	    		printf("\"%s --help\" for help.\n", argv[0]);
+                return( 1 );
+
             case 'x' :
 
-                sscanf( optarg, "%d", &opt.r_nbpps );
-                if( opt.r_nbpps < 1 || opt.r_nbpps > 1024 )
+                ret = sscanf( optarg, "%d", &opt.r_nbpps );
+                if( opt.r_nbpps < 1 || opt.r_nbpps > 1024 || ret != 1 )
                 {
-                    printf( "Invalid number of packets per second.\n" );
+                    printf( "Invalid number of packets per second. [1-1024]\n" );
+		    		printf("\"%s --help\" for help.\n", argv[0]);
                     return( 1 );
                 }
                 break;
@@ -822,6 +1037,7 @@ int main( int argc, char *argv[] )
                 if( getmac( optarg, 1, opt.r_bssid ) != 0 )
                 {
                     printf( "Invalid AP MAC address.\n" );
+		    		printf("\"%s --help\" for help.\n", argv[0]);
                     return( 1 );
                 }
                 break;
@@ -831,6 +1047,7 @@ int main( int argc, char *argv[] )
                 if( getmac( optarg, 1, opt.r_smac ) != 0 )
                 {
                     printf( "Invalid source MAC address.\n" );
+		    		printf("\"%s --help\" for help.\n", argv[0]);
                     return( 1 );
                 }
                 break;
@@ -840,24 +1057,28 @@ int main( int argc, char *argv[] )
                 if( opt.prga != NULL )
                 {
                     printf( "PRGA file already specified.\n" );
+		    		printf("\"%s --help\" for help.\n", argv[0]);
                     return( 1 );
                 }
                 if( opt.crypt != CRYPT_NONE )
                 {
                     printf( "Encryption key already specified.\n" );
+		    		printf("\"%s --help\" for help.\n", argv[0]);
                     return( 1 );
                 }
                 if( read_prga(&(opt.prga), optarg) != 0 )
                 {
+		    		printf("\"%s --help\" for help.\n", argv[0]);
                     return( 1 );
                 }
                 break;
 
             case 'i' :
 
-                if( opt.s_face != NULL )
+                if( opt.s_face != NULL || opt.s_file )
                 {
                     printf( "Packet source already specified.\n" );
+		    		printf("\"%s --help\" for help.\n", argv[0]);
                     return( 1 );
                 }
                 opt.s_face = optarg;
@@ -874,11 +1095,13 @@ int main( int argc, char *argv[] )
                 if( opt.prga != NULL )
                 {
                     printf( "PRGA file already specified.\n" );
+		    		printf("\"%s --help\" for help.\n", argv[0]);
                     return( 1 );
                 }
                 if( opt.crypt != CRYPT_NONE )
                 {
                     printf( "Encryption key already specified.\n" );
+		    		printf("\"%s --help\" for help.\n", argv[0]);
                     return( 1 );
                 }
 
@@ -896,6 +1119,7 @@ int main( int argc, char *argv[] )
                     if( n < 0 || n > 255 )
                     {
                         printf( "Invalid WEP key.\n" );
+			    		printf("\"%s --help\" for help.\n", argv[0]);
                         return( 1 );
                     }
 
@@ -918,22 +1142,86 @@ int main( int argc, char *argv[] )
                 if( i != 5 && i != 13 && i != 16 && i != 29 && i != 61 )
                 {
                     printf( "Invalid WEP key length.\n" );
+		    		printf("\"%s --help\" for help.\n", argv[0]);
                     return( 1 );
                 }
 
                 opt.weplen = i;
 
                 break;
+            case 'm':
+                if ( memcmp(opt.f_netmask, NULL_MAC, 6) != 0 )
+                {
+                    printf("Notice: netmask already given\n");
+		    		printf("\"%s --help\" for help.\n", argv[0]);
+                    break;
+                }
+                if(getmac(optarg, 1, opt.f_netmask) != 0)
+                {
+                    printf("Notice: invalid netmask\n");
+		    		printf("\"%s --help\" for help.\n", argv[0]);
+                    return( 1 );
+                }
+                break;
+            case 'd':
+                if ( memcmp(opt.f_bssid, NULL_MAC, 6) != 0 )
+                {
+                    printf("Notice: bssid already given\n");
+		    		printf("\"%s --help\" for help.\n", argv[0]);
+                    break;
+                }
+                if(getmac(optarg, 1, opt.f_bssid) != 0)
+                {
+                    printf("Notice: invalid bssid\n");
+		    		printf("\"%s --help\" for help.\n", argv[0]);
+                    return( 1 );
+                }
+                break;
+            case 'f':
+                opt.repeat = 1;
+                break;
+            case 'r' :
 
+                if( opt.s_face != NULL || opt.s_file )
+                {
+                    printf( "Packet source already specified.\n" );
+		    		printf("\"%s --help\" for help.\n", argv[0]);
+                    return( 1 );
+                }
+                opt.s_file = optarg;
+                break;
+
+            case 'H' :
+
+            	printf( usage, getVersion("Airtun-ng", _MAJ, _MIN, _SUB_MIN, _REVISION)  );
+            	return( 1 );
 
             default : goto usage;
         }
     }
 
-    if( argc - optind < 1 || argc - optind > 2 )
+    if( argc - optind != 1 )
     {
-    usage:
-        printf( usage, getVersion("Airtun-ng", _MAJ, _MIN, _SUB_MIN, _REVISION)  );
+    	if(argc == 1)
+    	{
+usage:
+	        printf( usage, getVersion("Airtun-ng", _MAJ, _MIN, _SUB_MIN, _REVISION)  );
+        }
+	    if( argc - optind == 0)
+	    {
+	    	printf("No replay interface specified.\n");
+	    }
+	    if(argc > 1)
+	    {
+    		printf("\"%s --help\" for help.\n", argv[0]);
+	    }
+        return( 1 );
+    }
+
+    if( ( memcmp(opt.f_netmask, NULL_MAC, 6) != 0 ) && ( memcmp(opt.f_bssid, NULL_MAC, 6) == 0 ) )
+    {
+        printf("Notice: specify bssid \"--bssid\" with \"--netmask\"\n");
+   		printf("\"%s --help\" for help.\n", argv[0]);
         return( 1 );
     }
 
@@ -948,12 +1236,14 @@ int main( int argc, char *argv[] )
     if( memcmp( opt.r_bssid, NULL_MAC, 6) == 0 )
     {
         printf( "Please specify a BSSID (-a).\n" );
+   		printf("\"%s --help\" for help.\n", argv[0]);
         return 1;
     }
 
     /* open the RTC device if necessary */
 
-#ifdef __i386__
+#if defined(__i386__)
+#if defined(linux)
     if( 1 )
     {
         if( ( dev.fd_rtc = open( "/dev/rtc", O_RDONLY ) ) < 0 )
@@ -982,10 +1272,12 @@ int main( int argc, char *argv[] )
             }
         }
     }
-#endif
+#endif /* linux */
+#endif /* __i386__ */
 
     /* create the RAW sockets */
 
+#if defined(linux)
     if( ( dev.fd_in = socket( PF_PACKET, SOCK_RAW,
                               htons( ETH_P_ALL ) ) ) < 0 )
     {
@@ -1055,7 +1347,7 @@ int main( int argc, char *argv[] )
         snprintf( strbuf,  sizeof( strbuf ) - 1,
                   "iwpriv %s rfmontx 1 >/dev/null 2>/dev/null",
                   argv[optind] );
-        system( strbuf );
+        unused = system( strbuf );
     }
 
     /* check if newer athXraw interface available */
@@ -1077,7 +1369,7 @@ int main( int argc, char *argv[] )
             memset( strbuf, 0, sizeof( strbuf ) );
             snprintf( strbuf,  sizeof( strbuf ) - 1,
                       "ifconfig %s up", athXraw );
-            system( strbuf );
+            unused = system( strbuf );
 
 #if 0 /* some people reported problems when prismheader is enabled */
             memset( strbuf, 0, sizeof( strbuf ) );
@@ -1093,6 +1385,31 @@ int main( int argc, char *argv[] )
         	dev.is_madwifing=1;
         }
     }
+#endif /* linux */
+
+#if defined(__FreeBSD__) || defined( __FreeBSD_kernel__)
+    for(i = 0;i < 10; i++) {
+        sprintf(buf, "/dev/bpf%d", i);
+
+        dev.fd_in = open(buf, O_RDWR);
+
+        if(dev.fd_in < 0) {
+            if(errno != EBUSY) {
+                perror("can't open /dev/bpf");
+                exit(1);
+            }
+            continue;
+        }
+        else
+            break;
+    }
+
+    if(dev.fd_in < 0) {
+        perror("can't open /dev/bpf");
+        exit(1);
+    }
+    dev.fd_out = dev.fd_in;
+#endif /* __FreeBSD__ */
 
     /* drop privileges */
 
@@ -1128,21 +1445,72 @@ int main( int argc, char *argv[] )
         dev.arptype_in = dev.arptype_out;
     }
 
+    if( opt.s_file != NULL )
+    {
+        if( ! ( dev.f_cap_in = fopen( opt.s_file, "rb" ) ) )
+        {
+            perror( "open failed" );
+            return( 1 );
+        }
+
+        n = sizeof( struct pcap_file_header );
+
+        if( fread( &dev.pfh_in, 1, n, dev.f_cap_in ) != (size_t) n )
+        {
+            perror( "fread(pcap file header) failed" );
+            return( 1 );
+        }
+
+        if( dev.pfh_in.magic != TCPDUMP_MAGIC &&
+            dev.pfh_in.magic != TCPDUMP_CIGAM )
+        {
+            fprintf( stderr, "\"%s\" isn't a pcap file (expected "
+                             "TCPDUMP_MAGIC).\n", opt.s_file );
+            return( 1 );
+        }
+
+        if( dev.pfh_in.magic == TCPDUMP_CIGAM )
+            SWAP32(dev.pfh_in.linktype);
+
+        if( dev.pfh_in.linktype != LINKTYPE_IEEE802_11 &&
+            dev.pfh_in.linktype != LINKTYPE_PRISM_HEADER )
+        {
+            fprintf( stderr, "Wrong linktype from pcap file header "
+                             "(expected LINKTYPE_IEEE802_11) -\n"
+                             "this doesn't look like a regular 802.11 "
+                             "capture.\n" );
+            return( 1 );
+        }
+    }
+
     dev.fd_tap = open( "/dev/net/tun", O_RDWR );
     if( dev.fd_tap < 0 )
     {
         printf( "error opening tap device: %s\n", strerror( errno ) );
+        printf( "try \"modprobe tun\"\n");
         return -1;
     }
     memset( &if_request, 0, sizeof( if_request ) );
+#if defined(linux)
     if_request.ifr_flags = IFF_TAP | IFF_NO_PI;
+#endif
     strncpy( if_request.ifr_name, "at%d", IFNAMSIZ );
+#if defined(linux)
     if( ioctl( dev.fd_tap, TUNSETIFF, (void *)&if_request ) < 0 )
     {
         printf( "error creating tap interface: %s\n", strerror( errno ) );
         close( dev.fd_tap );
         return -1;
     }
+#endif /* linux */
+#if defined(__FreeBSD__) || defined( __FreeBSD_kernel__)
+    if( ioctl( dev.fd_tap, SIOCGIFFLAGS, (void *)&if_request ) < 0 )
+    {
+        printf( "error creating tap interface: %s\n", strerror( errno ) );
+        close( dev.fd_tap );
+        return -1;
+    }
+#endif /* __FreeBSD__ */
     printf( "created tap interface %s\n", if_request.ifr_name );
 
     if(opt.prgalen <= 0 && opt.crypt == CRYPT_NONE)
@@ -1169,6 +1537,80 @@ int main( int argc, char *argv[] )
 
     for( ; ; )
     {
+        if(opt.s_file != NULL)
+        {
+            n = sizeof( pkh );
+
+            if( fread( &pkh, n, 1, dev.f_cap_in ) != 1 )
+            {
+                printf("Finished reading input file %s.\n", opt.s_file);
+                opt.s_file = NULL;
+                continue;
+            }
+
+            if( dev.pfh_in.magic == TCPDUMP_CIGAM )
+                SWAP32( pkh.caplen );
+
+            n = caplen = pkh.caplen;
+
+            if( n <= 0 || n > (int) sizeof( h80211 ) )
+            {
+                printf("Finished reading input file %s.\n", opt.s_file);
+                opt.s_file = NULL;
+                continue;
+            }
+
+            if( fread( h80211, n, 1, dev.f_cap_in ) != 1 )
+            {
+                printf("Finished reading input file %s.\n", opt.s_file);
+                opt.s_file = NULL;
+                continue;
+            }
+
+            if( dev.pfh_in.linktype == LINKTYPE_PRISM_HEADER )
+            {
+                if( h80211[7] == 0x40 )
+                    n = 64;
+                else
+                    n = *(int *)( h80211 + 4 );
+
+                if( n < 8 || n >= (int) caplen )
+                    continue;
+
+                memcpy( tmpbuf, h80211, caplen );
+                caplen -= n;
+                memcpy( h80211, tmpbuf + n, caplen );
+            }
+
+            if( opt.repeat )
+            {
+                if( memcmp(opt.f_bssid, NULL_MAC, 6) != 0 )
+                {
+                    switch( h80211[1] & 3 )
+                    {
+                        case  0: memcpy( bssid, h80211 + 16, 6 ); break;
+                        case  1: memcpy( bssid, h80211 +  4, 6 ); break;
+                        case  2: memcpy( bssid, h80211 + 10, 6 ); break;
+                        default: memcpy( bssid, h80211 +  4, 6 ); break;
+                    }
+                    if( memcmp(opt.f_netmask, NULL_MAC, 6) != 0 )
+                    {
+                        if(is_filtered_netmask(bssid)) continue;
+                    }
+                    else
+                    {
+                        if( memcmp(opt.f_bssid, bssid, 6) != 0 ) continue;
+                    }
+                }
+                send_packet(h80211, caplen);
+            }
+
+
+            packet_recv( h80211, caplen);
+            msleep( 1000/opt.r_nbpps );
+            continue;
+        }
+
         FD_ZERO( &read_fds );
         FD_SET( dev.fd_in, &read_fds );
         FD_SET( dev.fd_tap, &read_fds );
